#include "iplookup.h"


trie_t *create_trie (struct net_table *in_table, ip_addr_t *tot_nodes) {
  struct net_table *table;  /*  array version  */
  ip_addr_t i, tot_leaves = 0, tmp;
  trie_t *root;
  printf ("Creating trie...");
  table = in_table;
  /*  Create root for no route to host  */
  if ((root = create_root ()) == NULL) {
      printf ("Memory allocation failure...can't go on\nProgram aborted\n");
      return NULL;
  }
  tot_leaves = 2;  /*  root was partitioning in tow segments  */
  for (i = 0; table != NULL; i ++) {
    tmp = trie_insert_network (table, root);
    table = table->next;
    /*  Dispose in_table entry  */
    free (in_table);
    in_table = table;
    tot_leaves += tmp;
    if (tmp == -1) {
      printf ("Memory allocation failure...can't go on\nProgram aborted\n");
      return NULL;
    }
  }
  printf ("done (%u nodes)\n", tot_leaves);
  *tot_nodes = tot_leaves;
  return root;
}

trie_t *create_root (void) {
  int i;
  trie_t *tmp, *root;
  root = (trie_t *) malloc (sizeof (trie_t));
  if (!root) return NULL;
  root->mark = 0;
  root->next_hop = 0; /*  useless value in root  */
  for (i = 0; i < 2; i ++) {
    root->child[i] = malloc (sizeof (trie_t));
    if (!root->child[i]) {
      if (i == 1) free (root->child[0]);
      free (root);
      return NULL;
    }
    tmp = root->child[i];
    tmp->mark = 1;
    tmp->next_hop = 0;   /*  NO ROUTE TO HOST  */
    tmp->child[0] = NULL;
    tmp->child[1] = NULL;
  }
  return root;
}

rle_t **build_RLE_sequence (rle_t **RLE_seq, ip_addr_t *RLE_len) {
  ip_addr_t i,j;
  j = 0;
  for (i = 1; i < *RLE_len; i ++)
    if (RLE_seq[j]->next_hop != RLE_seq[i]->next_hop) {
      RLE_seq[j]->run_length = RLE_seq[i]->run_length - RLE_seq[j]->run_length;
      j ++;
      RLE_seq[j]->next_hop = RLE_seq[i]->next_hop;
      RLE_seq[j]->run_length = RLE_seq[i]->run_length;
    }
  /*  Add last element  */
  RLE_seq[j]->run_length = ~0 - RLE_seq[j]->run_length + 1;
  j ++;
  /*  free useles resources and adjust RLE_seq size  */
  for (i = j; i < *RLE_len; i ++)
    free (RLE_seq[i]);
  *RLE_len = j;
  return (rle_t **) realloc (RLE_seq, sizeof (rle_t *) * j);
}


rle_t **build_empty_RLE_sequence (ip_addr_t len) {
  ip_addr_t i;
  rle_t **RLE;
  RLE = (rle_t **) malloc (len * sizeof (rle_t *));
  if (RLE == NULL) return NULL;
  for (i = 0; i < len; i ++) {
    RLE[i] = (rle_t *) malloc (sizeof (rle_t));
    if (RLE[i] == NULL) {
      ip_addr_t j;
      for (j = 0; j < i; j ++) free (RLE[j]);
      free (RLE);
      return NULL;
    }
  }
  return RLE;
}

int trie_remove_network (ip_addr_t ip, ip_t netmask, trie_t *root) {
  ip_t i;
  int kid;
  trie_t *node, *father, *ancestor;
  node = root;
  father = root;
  ancestor = root;
  for (i = 0; i < netmask - 1; i ++) {
    if (node->mark == 1) ancestor = node;
    kid = bit_k (ip, 31 - i);
    node = node->child[kid];
    if (node == NULL) {
      printf ("Route not exist...delete ignored\n");
      return -1;
    }
  }
  kid = bit_k (ip, 31 - i);
  father = node;
  node = node->child[kid];
  if (node->mark == 0) {
    printf ("Route not exist...delete ignored\n");
    return -1;
  }
  node->mark = 0;
  node->next_hop = father->next_hop;
  cover_holes (node);
  return merge_leaves (ancestor, 0);
}

ip_addr_t merge_leaves (trie_t *root, ip_addr_t delete) {
  trie_t *child[2];
  char leaf = 0;
  int i;
  /* the trie is complete, root have no child or two children  */
  if (!root->child[0] && !root->child[1]) return delete;
  if (root->child[0]) delete = merge_leaves (root->child[0], delete);
  if (root->child[1]) delete = merge_leaves (root->child[1], delete);
  for (i = 0; i < 2; i ++) {
    child[i] = root->child[i];
    /* is root of tree with depth 1  */
    if (!child[i]->child[0] && !child[i]->child[1]) leaf ++;
  }
  if (leaf == 2) {
    if (child[0]->next_hop == child[1]->next_hop) {
      free (child[0]);
      free (child[1]);
      root->child[0] = NULL;
      root->child[1] = NULL;
      return delete + 1;
    }
  }
  return delete;
}

int trie_insert_network (struct net_table *elem, trie_t *root) {
  ip_t i, tot_leaves = 0;
  int kid, other;
  trie_t *node, *tmp;
  node = root;
  for (i = 0; i < elem->netmask; i ++) {
    kid = bit_k (elem->ip, 31 - i);
    other = kid == 0 ? 1 : 0;
    if (node->child[kid] == NULL) {   /*  Haven't chlid from kid side  */
      node->child[kid] = malloc (sizeof (trie_t));
      if (!node->child[kid]) return -1;
      tmp = (trie_t *) node->child[kid];
      tmp->next_hop = node->next_hop;
      tmp->child[0] = NULL;
      tmp->child[1] = NULL;
      tmp->mark = 0;
      node->child[other] = malloc (sizeof (trie_t));
      if (!node->child[other]) return -1;
      tmp = (trie_t *) node->child[other];
      tmp->next_hop = node->next_hop;
      tmp->child[0] = NULL;
      tmp->child[1] = NULL;
      tmp->mark = 0;
      tot_leaves ++;
    }
    node = node->child[kid];
  }
  node->next_hop = elem->next_hop;
  node->mark = 1;
  cover_holes (node);
  return tot_leaves;
}

void cover_holes (trie_t *root) {
  if (root->child[0] != NULL) {
    if (((trie_t*) root->child[0])->mark == 0) {
      ((trie_t*) root->child[0])->next_hop = root->next_hop;
      cover_holes (root->child[0]);
    }
  }
  if (root->child[1] != NULL) {
    if (((trie_t*) root->child[1])->mark == 0) {
      ((trie_t*) root->child[1])->next_hop = root->next_hop;
      cover_holes (root->child[1]);
    }
  }
}

void cover_holes_old (trie_t *root) {
  trie_t *tmp;
  int i;
  for (i = 0; i < 2; i ++) {
    if (root->child[i] == NULL) return;
    tmp = root->child[i];
    if (tmp->mark == 0) {
      tmp->next_hop = root->next_hop;
      cover_holes (tmp);
    }
  }
}

rle_t **update_RLE_sequence (trie_t *root, ip_addr_t *RLE_len, char flag) {
  rle_t **RLE_seq;
  printf ("Creating RLE sequece...");
  RLE_seq = build_empty_RLE_sequence (*RLE_len);
  if (RLE_seq == NULL) {
    printf ("fail\n");
    return NULL;
  }
  *RLE_len = 0;
  switch (flag) {
  case 0:  /* Crea la sequenza sup */
    visit_b_tree_sup (root, RLE_seq, RLE_len, 0, 0);
    break;
  case 1:  /* Crea la sequenza inf */
    visit_b_tree_inf (root, RLE_seq, RLE_len, 0, 0);
    break;
  default:
    visit_b_tree (root, RLE_seq, RLE_len, 0, 0);
    break;
  }
  RLE_seq = build_RLE_sequence (RLE_seq, RLE_len);
  printf ("done\n");
  return RLE_seq;
}

void visit_b_tree_sup (trie_t *root, rle_t **RLE, ip_addr_t *RLE_ind, ip_addr_t ip, ip_t netmask) {
  /*  if is a leaf  */
  if (root->child[0] == NULL && root->child[1] == NULL) {
    ip = (ip << (32 - netmask));
    RLE[*RLE_ind]->next_hop = root->next_hop;
    RLE[*RLE_ind]->run_length = ip;
    *RLE_ind = *RLE_ind + 1;
    return;
  }
  /*  Netmask is 24, but is not a leaf  */
  if (netmask == 24) {
    ip = (ip << (32 - netmask));
    RLE[*RLE_ind]->next_hop = FOLLOW_UP;
    RLE[*RLE_ind]->run_length = ip;
    *RLE_ind = *RLE_ind + 1;
    return;
  }
  
  if (root->child[0] != NULL)
    visit_b_tree_sup (root->child[0], RLE, RLE_ind, (ip << 1), netmask + 1);
  if (root->child[1] != NULL)
    visit_b_tree_sup (root->child[1], RLE, RLE_ind, (ip << 1)+1, netmask+1);
}

void visit_b_tree_inf (trie_t *root, rle_t **RLE, ip_addr_t *RLE_ind, ip_addr_t ip, ip_t netmask) {
  /*  if is a leaf  */
  if (root->child[0] == NULL && root->child[1] == NULL && netmask > 24) {
    ip = (ip << (32 - netmask));
    RLE[*RLE_ind]->next_hop = root->next_hop;
    RLE[*RLE_ind]->run_length = ip;
    *RLE_ind = *RLE_ind + 1;
    return;
  }
  /*  Only for initial case  */  
  if (root->child[0] == NULL && root->child[1] == NULL && ip == 0) {
    ip = (ip << (32 - netmask));
    RLE[*RLE_ind]->next_hop = root->next_hop;
    RLE[*RLE_ind]->run_length = ip;
    *RLE_ind = *RLE_ind + 1;
    return;
  }
  if (root->child[0] != NULL)
    visit_b_tree_inf (root->child[0], RLE, RLE_ind, (ip << 1), netmask + 1);
  if (root->child[1] != NULL)
    visit_b_tree_inf (root->child[1], RLE, RLE_ind, (ip << 1)+1, netmask+1);
}

void visit_b_tree (trie_t *root, rle_t **RLE, ip_addr_t *RLE_ind, ip_addr_t ip, ip_t netmask) {
  /*  if is a leaf  */
  if (root->child[0] == NULL && root->child[1] == NULL) {
    ip = (ip << (32 - netmask));
    RLE[*RLE_ind]->next_hop = root->next_hop;
    RLE[*RLE_ind]->run_length = ip;
    *RLE_ind = *RLE_ind + 1;
    return;
  }
  
  if (root->child[0] != NULL)
    visit_b_tree (root->child[0], RLE, RLE_ind, (ip << 1), netmask + 1);
  if (root->child[1] != NULL)
    visit_b_tree (root->child[1], RLE, RLE_ind, (ip << 1)+1, netmask+1);
}

